/// \file
///
/// This file is part of RakNet Copyright 2003 Kevin Jenkins.
///
/// Usage of Raknet is subject to the appropriate licence agreement.
/// "Shareware" Licensees with Rakkarsoft LLC are subject to the
/// shareware license found at
/// http://www.rakkarsoft.com/shareWareLicense.html
/// "Commercial" Licensees are subject to the commercial license found at
/// http://www.rakkarsoft.com/sourceCodeLicense.html
/// Custom license users are subject to the terms therein.
/// All other users are subject to the GNU General Public
/// License as published by the Free
/// Software Foundation; either version 2 of the License, or (at your
/// option) any later version.

#include "DataReplicator.h"
#include "PacketEnumerations.h"
#include "RakPeerInterface.h"
#include "GetTime.h"
#include "StringCompressor.h"
#include <string.h>
#include <stdio.h>
#include <assert.h>

strToIndexMapType DataReplicator::memoryToStringMappingKey=0;

int BaseObjectDataComp( DataReplicator::BaseObjectData* data, ReplicatedObject* key )
{
	if (key < data->replicatedObject)
		return -1;
	if (key==data->replicatedObject)
		return 0;
	return 1;
}

int BaseMemoryDataComp( DataReplicator::BaseMemoryData* data, strToIndexMapType key )
{
	if (key < data->localKey)
		return -1;
	if (key==data->localKey)
		return 0;
	return 1;
}

int ExtendedMemoryDataComp( DataReplicator::ExtendedMemoryData* data, DataReplicator::BaseData * key )
{
	if (key < data->baseData)
		return -1;
	if (key==data->baseData)
		return 0;
	return 1;
}

int ExtendedObjectDataComp( DataReplicator::ExtendedObjectData* data, DataReplicator::BaseData * key )
{
	if (key < data->baseData)
		return -1;
	if (key==data->baseData)
		return 0;
	return 1;
}

int ParticipantStructComp( DataReplicator::ParticipantStruct* data, PlayerID key )
{
	if (key < data->playerId)
		return -1;
	if (key==data->playerId)
		return 0;
	return 1;
}
void DefaultSerializeCB(SerializeCallbackParameters *parms)
{
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
	parms->outBitStream->Write((char*)parms->sourceMemory, parms->memoryByteLength);
}
void DefaultDeserializeCB(DeserializeCallbackParameters *parms)
{
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
	parms->inBitStream->Read((char*)parms->destinationMemory, parms->memoryByteLength);
}
void* DefaultMakeCopyCB(void *memory, unsigned memoryByteLength, void *context)
{
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif // warning C4100: <variable name> : unreferenced formal parameter
	char *copy = new char[memoryByteLength];
	memcpy(copy, memory, memoryByteLength);
	return (void*)copy;
}
void DefaultCopyFromCB(void *destination, void *source, unsigned memoryByteLength, void *context)
{
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
	memcpy(destination, source, memoryByteLength);
}
void DefaultFreeMemoryCB(void *memory,unsigned byteLength, void *context)
{
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
	delete (unsigned char*)memory;
}
bool DefaultShouldSendUpdateCB(ShouldSendUpdateCallbackParameters *parms)
{
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
	if (parms->lastSentValue==0)
		return true;
	return memcmp(parms->memory, parms->lastSentValue, parms->memoryByteLength)!=0;
}
DataReplicator::DataReplicator()
{
	rakPeer=0;
	objectFactoryCallback=0;
	objectCreatedFromNetwork=ignoreReplicateObjectCalls=false;
	broadcastToAll=false;
	enableReplication=true;

	localCreationRequestIndex=0;
	memset(localCreationRequestTable, 0, sizeof(NetworkIDGenerator*)*LOCAL_CREATION_REQUEST_TABLE_SIZE);
}

DataReplicator::~DataReplicator()
{
	Clear();
}
void DataReplicator::SetBroadcastToAll(bool value)
{
	broadcastToAll=value;
}

bool DataReplicator::GetBroadcastToAll(void) const
{
	return broadcastToAll;
}
void DataReplicator::AddParticipant(PlayerID playerId)
{
	if (playerId==UNASSIGNED_PLAYER_ID)
	{
		unsigned short peerIndex;

		for (peerIndex=0; peerIndex < rakPeer->GetMaximumNumberOfPeers(); peerIndex++)
		{
			playerId=rakPeer->GetPlayerIDFromIndex(peerIndex);
			if (playerId!=UNASSIGNED_PLAYER_ID)
				AddToParticipantList(playerId);
		}
	}
	else
	{
		AddToParticipantList(playerId);
	}
}
void DataReplicator::RemoveParticipant(PlayerID playerId)
{
	if (playerId==UNASSIGNED_PLAYER_ID)
	{
		RemoveAllParticipants();
	}
	else
	{
		RemoveFromParticipantList(playerId, false, false);
	}
}
void DataReplicator::EnableReplication(bool value)
{
	enableReplication=value;
}
void DataReplicator::SynchronizeMemory(const char *stringId, void *memory, unsigned memoryByteLength, bool isMemoryOwner, int maxUpdateFrequencyMS, ReplicatedObject *replicatedObject, PacketPriority priority, PacketReliability reliability, char orderingChannel)
{
	BaseMemoryData *baseData;
	unsigned i;

	if (replicatedObject!=0 && GetBaseObjectByReplicatedObject(replicatedObject)==0)
	{
		assert(0); // Unknown object
		return;
	}

	char newStringId[256];
	assert(strlen(stringId) < 246);

	if (replicatedObject==0)
		strcpy(newStringId, stringId);
	else
	{
#ifdef _DEBUG
		// You shouldn't synchronize memory if you are waiting for the object creation to be verified from the server
		assert(replicatedObject->GetWaitingForCreationResponse()==false);
#endif
		sprintf(newStringId, "%s__%i", stringId, replicatedObject->GetNetworkID());
	}
	bool usedToSend;
	bool usedToReceive;
	bool nowSend;
	bool nowReceive;

	if ((baseData=GetBaseMemoryDataByString(stringId))!=0)
	{
		usedToSend=baseData->isOwner || (baseData->associatedObject && baseData->associatedObject->IsNetworkIDAuthority());
		usedToReceive=baseData->isOwner==false || (baseData->associatedObject && baseData->associatedObject->IsNetworkIDAuthority()==false);
		nowSend=isMemoryOwner || (baseData->associatedObject && baseData->associatedObject->IsNetworkIDAuthority());;
		nowReceive=isMemoryOwner==false || (baseData->associatedObject && baseData->associatedObject->IsNetworkIDAuthority()==false);

		// If used to receive and now don't receive, tell everyone to stop sending to us
		if ( usedToReceive && nowReceive==false )
		{
			for (i=0; i < participantList.Size(); i++)
				SendMemoryStopRequest(baseData->localIdentifier, participantList[i]->playerId);
		}

		// Already synchronized
		baseData->associatedObject=replicatedObject;
		baseData->synchronizedMemory=memory;
		//baseData->isUnifiedMemory=isUnifiedMemory;
		baseData->isOwner=isMemoryOwner;
		baseData->priority=priority;
		baseData->reliability=reliability;
		baseData->orderingChannel=orderingChannel;
		baseData->maxUpdateFrequencyMS=maxUpdateFrequencyMS;
		baseData->gotFirstUpdate=false;
	}
	else
	{
		usedToSend=false;
		usedToReceive=false;

		nowSend=isMemoryOwner || (replicatedObject && replicatedObject->IsNetworkIDAuthority());;
		nowReceive=isMemoryOwner==false || (replicatedObject && replicatedObject->IsNetworkIDAuthority()==false);

		baseData = new BaseMemoryData;
		// Base class stuff
		baseData->priority=priority;
		baseData->reliability=reliability;
		baseData->orderingChannel=orderingChannel;
		baseData->maxUpdateFrequencyMS=maxUpdateFrequencyMS;
		baseData->lastUpdateTime=0;
		baseData->lastReceivedTime=0;
		usedToSend=false;		

		// Extended stuff
		baseData->localIdentifier=new char [strlen(newStringId)+1];
		strcpy(baseData->localIdentifier, newStringId);
		baseData->localKey=GenerateMemoryStringMappingKey();
		baseData->associatedObject=replicatedObject;
		baseData->synchronizedMemory=memory;
		//baseData->isUnifiedMemory=isUnifiedMemory;
		baseData->serializeCB=DefaultSerializeCB;
		baseData->deserializeCB=DefaultDeserializeCB;
		baseData->makeCopyCB=DefaultMakeCopyCB;
		baseData->copyFromCB=DefaultCopyFromCB;
		baseData->freeMemoryCB=DefaultFreeMemoryCB;
		baseData->shouldSendUpdateCB=DefaultShouldSendUpdateCB;
		baseData->byteLength=memoryByteLength;
		baseData->context=0;
		baseData->lastInterpolateCallTime=0;
		baseData->interpolateCB=0;
		//baseData->calculateExtrapolatedValueCB=0;
		baseData->lastReceivedValue=0;
		baseData->gotFirstUpdate=false;
		//	baseData->lastExtrapolatedValue=0;
		baseData->isOwner=isMemoryOwner;
		baseData->needsInterpolation=false;
		baseData->userInterpolationFinished=false;
		baseData->lastNonInterpValue=0;

		// Add to the list
		memoryList.InsertElement(baseData, baseData->localKey, BaseMemoryDataComp);
	}

	if (usedToSend==false && nowSend==true)
	{
		unsigned i;
		for (i=0; i < participantList.Size(); i++)
		{
			AddExtendedMemoryDataToParticipant(baseData, participantList[i]);
		}
	}
}
/*
void* DataReplicator::GetLastSentMemoryValue(const void *memory, PlayerID playerId)
{
BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memory);
if (baseData)
{
ExtendedMemoryData *extendedMemoryData = GetExtendedMemoryByBaseMemory(baseData, GetParticipantByPlayerID(playerId));
if (extendedMemoryData)
return extendedMemoryData->lastMemorySend;
}
return 0;
}
*/
void* DataReplicator::GetLastSentMemoryValue(const void *memoryPtr, PlayerID playerId)
{
	/*
	BaseMemoryData *baseData;
	unsigned i;
	baseData=0;
	for (i=0; i < memoryList.Size(); i++)
	{
	if (memoryList[i]->synchronizedMemory->GetType()==SYNCHRONIZED_MEMORY_TYPE_DEFAULT && ((SynchronizedMemoryDefault*)(memoryList[i]->synchronizedMemory))->GetUserData()==memoryPtr)
	{
	baseData=memoryList[i];
	break;
	}
	}

	if (baseData)
	{
	ExtendedMemoryData *extendedMemoryData = GetExtendedMemoryByBaseMemory(baseData, GetParticipantByPlayerID(playerId));
	if (extendedMemoryData)
	return ((SynchronizedMemoryDefault*)(extendedMemoryData->lastMemorySend))->GetUserData();
	}
	*/
	ParticipantStruct *participantStruct = GetParticipantByPlayerID(playerId);
	BaseMemoryData *baseMemoryData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
	if (baseMemoryData && participantStruct)
	{
		ExtendedMemoryData *extendedMemoryData = GetExtendedMemoryByBaseMemory(baseMemoryData, participantStruct);
		if (extendedMemoryData)
			return extendedMemoryData->lastMemorySend;
	}

	return 0;
}
unsigned int DataReplicator::GetLastSendTime(const void *memoryPtr, PlayerID playerId)
{
	ParticipantStruct *participantStruct = GetParticipantByPlayerID(playerId);
	BaseMemoryData *baseMemoryData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
	if (baseMemoryData && participantStruct)
	{
		ExtendedMemoryData *extendedMemoryData = GetExtendedMemoryByBaseMemory(baseMemoryData, participantStruct);
		if (extendedMemoryData)
			return extendedMemoryData->lastSendTime;
	}

	return 0;
}
/*
void DataReplicator::UnsynchronizeMemory(const char *memoryPtr, bool sendNotificationPacket)
{
BaseMemoryData *baseData;
unsigned i;
baseData=0;
for (i=0; i < memoryList.Size(); i++)
{
if (memoryList[i]->synchronizedMemory->GetType()==SYNCHRONIZED_MEMORY_TYPE_DEFAULT && ((SynchronizedMemoryDefault*)(memoryList[i]->synchronizedMemory))->GetUserData()==memoryPtr)
{
baseData=memoryList[i];
break;
}
}

if (baseData)
UnsynchronizeMemoryInt(baseData, sendNotificationPacket);
}
*/
void DataReplicator::UnsynchronizeMemory(const void *memory, bool sendNotificationPacket)
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memory);
	if (baseData)
		UnsynchronizeMemoryInt(baseData, sendNotificationPacket);
}
void DataReplicator::SetMemoryContext(const void *memoryPtr, void *_context)
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
#endif
	baseData->context=_context;
}
void DataReplicator::InterpolateMemory(const void *memoryPtr,
						  bool (* _interpolateCB)(void *currentValue, void *targetValue, unsigned int currentTimeMS, unsigned int timePacketSentMS, unsigned int lastCallTimeMS, void *context)//,
						  //void (* _calculateExtrapolatedValueCB)(void *extrapolatedValue, void *lastReceivedValue, unsigned int currentTimeMS, unsigned int timePacketSentMS, void *context )
						  )
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
	assert(_interpolateCB);
#endif
    baseData->interpolateCB=_interpolateCB;
	//baseData->calculateExtrapolatedValueCB=_calculateExtrapolatedValueCB;
}
void DataReplicator::SetMemorySerializeCB(const void *memoryPtr, void (* cbFunc)(SerializeCallbackParameters *parms) )
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
	assert(cbFunc);
#endif
	baseData->serializeCB=cbFunc;
}
void DataReplicator::SetMemoryDeserializeCB(const void *memoryPtr, void (* cbFunc)(DeserializeCallbackParameters *parms))
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
	assert(cbFunc);
#endif
	baseData->deserializeCB=cbFunc;
}
void DataReplicator::SetMemoryMakeCopyCB(const void *memoryPtr, void* (* cbFunc)(void *source, unsigned memoryByteLength, void *context))
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
	assert(cbFunc);
#endif
	baseData->makeCopyCB=cbFunc;
}
void DataReplicator::SetMemoryCopyFromCB(const void *memoryPtr, void (* cbFunc)(void *destination, void *source, unsigned memoryByteLength, void *context))
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
	assert(cbFunc);
#endif
	baseData->copyFromCB=cbFunc;
}
void DataReplicator::SetMemoryFreeMemoryCB(const void *memoryPtr, void (* cbFunc)(void *memory, unsigned memoryByteLength, void *context))
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
	assert(cbFunc);
#endif
	baseData->freeMemoryCB=cbFunc;
}
void DataReplicator::SetMemoryShouldSendUpdateCB(const void *memoryPtr, bool (* cbFunc)(ShouldSendUpdateCallbackParameters *parms))
{
	BaseMemoryData *baseData = GetBaseMemoryDataBySynchronizedMemory(memoryPtr);
#ifdef _DEBUG
	assert(baseData);
	assert(cbFunc);
#endif
	baseData->shouldSendUpdateCB=cbFunc;
}
void DataReplicator::RequestObjectCreation(const char *objectName, RakNet::BitStream *context, ReplicatedObject *object, PacketPriority priority, PacketReliability reliability, char orderingChannel, PlayerID targetSystem)
{
	assert(objectName || object);

	// If this assert hits you are trying to request an object you cannot create.  Be sure that for one of your classes,
	// ReplicatedObject::GetClassName returns the exact string passed here.
	//assert(ObjectReplicationHandler::GetHandler(objectName));

	RakNet::BitStream outBitStream;
	PlayerID ownerId;

	if (targetSystem!=UNASSIGNED_PLAYER_ID)
	{
		ParticipantStruct *participantStruct=GetParticipantByPlayerID(targetSystem);
		if (participantStruct==0)
		{
#ifdef _DEBUG
			// If this hits then you are sending to someone that is not a participant in the data replication system, which is not allowed.
			// Either add this player as a participant or call AddParticipant(UNASSIGNED_PLAYER_ID) to allow sends to all.
			assert(0);
#endif
			return;
		}

		participantStruct->objectRequestCount++;

		outBitStream.Write((unsigned char) ID_REPLICATOR_OBJECT_CREATION_REQUEST);
		EncodeClassName(object ? object->GetObjectClassName() : objectName, targetSystem, &outBitStream);

		if (object)
		{
		//	outBitStream.Write(true);
			localCreationRequestTable[localCreationRequestIndex]=object;
			outBitStream.Write(localCreationRequestIndex);
			if (++localCreationRequestIndex==LOCAL_CREATION_REQUEST_TABLE_SIZE)
				localCreationRequestIndex=0;
		}
		else
		{
		//	outBitStream.Write(false);
		}

		ownerId=object->GetOwner();
		if (ownerId==rakPeer->GetInternalID())
			ownerId=rakPeer->GetExternalID(targetSystem);

		// Write owner
		outBitStream.Write(ownerId.binaryAddress);
		outBitStream.Write(ownerId.port);

		if (context)
			outBitStream.Write(context);

		rakPeer->Send(&outBitStream, priority, RELIABLE_ORDERED, 0, targetSystem, false);
	}
	else
	{
		// Send to all participants
		unsigned i;
		for (i=0; i < participantList.Size(); i++)
		{
			outBitStream.Reset();
			outBitStream.Write((unsigned char) ID_REPLICATOR_OBJECT_CREATION_REQUEST);
			EncodeClassName(object ? object->GetObjectClassName() : objectName, participantList[i]->playerId, &outBitStream);

			if (object)
			{
			//	outBitStream.Write(true);
				localCreationRequestTable[localCreationRequestIndex]=object;
				outBitStream.Write(localCreationRequestIndex);
			}
			else
			{
		//		outBitStream.Write(false);
			}

			if (context)
				outBitStream.Write(context);

			ownerId=object->GetOwner();
			if (ownerId==rakPeer->GetInternalID())
				ownerId=rakPeer->GetExternalID(targetSystem);

			// Write owner
			outBitStream.Write(ownerId.binaryAddress);
			outBitStream.Write(ownerId.port);

			rakPeer->Send(&outBitStream, priority, RELIABLE_ORDERED, 0, participantList[i]->playerId, false);

			// Record that we sent a packet asking for an object to be created.  We don't record what particular object we ask for
			// because the recipient can change that.
			participantList[i]->objectRequestCount++;
		}


		if (object)
			if (++localCreationRequestIndex==LOCAL_CREATION_REQUEST_TABLE_SIZE)
				localCreationRequestIndex=0;
	}
}


void DataReplicator::SetClassFactory(ReplicatedObject* (* factoryFunc)(const char *className, RakNet::BitStream *inContext))
{
	objectFactoryCallback=factoryFunc;
}

ReplicatedObject* DataReplicator::ReplicateObject(const char *objectName, ReplicatedObject *object, bool isObjectOwner, PacketPriority priority, PacketReliability reliability, char orderingChannel, int maxUpdateFrequencyMS)
{
#ifdef _DEBUG
	assert(object || objectName);
#endif

	if (ignoreReplicateObjectCalls)
		return object;

	unsigned i;
	ReplicatedObject *objectToReplicate;

	if (object)
	{
		for (i=0; i < objectList.Size(); i++)
			if (objectList[i]->replicatedObject==object)
				return object; // Object already registered

		objectToReplicate=object;
	}
	else
	{
		ignoreReplicateObjectCalls=true;
		objectToReplicate=CreateObject(objectName, 0);
		ignoreReplicateObjectCalls=false;
		if (objectToReplicate==0)
			return 0;
	}

	BaseObjectData *baseData = new BaseObjectData;
	// Base class stuff
	baseData->priority=priority;
	baseData->reliability=reliability;
	baseData->orderingChannel=orderingChannel;
	baseData->maxUpdateFrequencyMS=maxUpdateFrequencyMS;
	baseData->lastUpdateTime=0;

	// Extended stuff
	baseData->replicatedObject=objectToReplicate;
	baseData->calledOnReplication=false;
	baseData->isOwner=isObjectOwner;

	// Add to the list
	objectList.InsertElement(baseData, baseData->replicatedObject, BaseObjectDataComp);

	// If we own the object, or are the authority, then we send.  So save the per-system data for this object
	if (baseData->isOwner || baseData->replicatedObject->IsNetworkIDAuthority())
	{
		for (i=0; i < participantList.Size(); i++)
		{
			AddExtendedObjectDataToParticipant(baseData, participantList[i]);
		}
	}

	if (isObjectOwner)
		objectToReplicate->SetOwner(rakPeer->GetInternalID());

	// Don't call OnReplication until the object is accepted because we need a object id to synch memory
//	objectToReplicate->OnReplication(this);

	// objectCreatedFromNetwork is used so we know if we are currently processing a creation packet.  We otherwise wouldn't know, because the
	// packet handling code calls callbacks (include the object constructor) which could in turn call this function.
	// This keeps us from sending an object creation request to the system that just sent us an object to create.
	if (objectToReplicate->IsNetworkIDAuthority()==false && objectCreatedFromNetwork==false)
	{
		assert(objectToReplicate->GetNetworkID()==UNASSIGNED_OBJECT_ID);

		// We are not the owner, so request object creation or replication
		if (objectToReplicate)
			objectName=objectToReplicate->GetObjectClassName();

		ParticipantStruct *participantStruct=GetParticipantByPlayerID(objectToReplicate->GetNetworkIDAuthority());
		if (participantStruct==0)
		{
		#ifdef _DEBUG
			// If this hits then you are sending to someone that is not a participant in the data replication system, which is not allowed.
			// Either add this player as a participant or call AddParticipant(UNASSIGNED_PLAYER_ID) to allow sends to all.
			assert(0);
		#endif
			return objectToReplicate;
		}

		baseData->needsCreationRequest=true;
	}
	else
	{
		baseData->needsCreationRequest=false;
		// Defer replication call until update loop in case this function is called by an object constructor, in which case no network ID has been assigned yet.
	}

	return objectToReplicate;
}

void DataReplicator::AddExtendedMemoryDataToParticipant(BaseMemoryData *baseMemoryData, ParticipantStruct *participantStruct)
{
	ExtendedMemoryData *extendedData = new ExtendedMemoryData;
	extendedData->remoteKey=STR_TO_INDEX_MAP_MAX;
	extendedData->baseData=baseMemoryData;
	extendedData->lastMemorySend=0;
	extendedData->lastSendTime=0;
	participantStruct->extendedMemoryList.InsertElement(extendedData, extendedData->baseData, ExtendedMemoryDataComp);
}

void DataReplicator::AddExtendedObjectDataToParticipant(BaseObjectData *baseObjectData, ParticipantStruct *participantStruct)
{
	assert(baseObjectData->isOwner || baseObjectData->replicatedObject->IsNetworkIDAuthority());

	ExtendedObjectData *extendedData = new ExtendedObjectData;
	// Base class stuff
	extendedData->baseData=(BaseData*)baseObjectData;

	// Extended stuff
    extendedData->objectInstantiated=false;
	extendedData->objectInScope=false;

	participantStruct->extendedObjectList.InsertElement(extendedData, extendedData->baseData, ExtendedObjectDataComp);
}
bool DataReplicator::ObjectInScope(ReplicatedObject *object, PlayerID target)
{
	ParticipantStruct *participant = GetParticipantByPlayerID(target);
	if (participant==0)
		return false;
	ExtendedObjectData *extendedObjectData = GetExtendedObjectByReplicatedObject(object, participant);
	if (extendedObjectData)
		return extendedObjectData->objectInScope;
	else
		return true; // If we don't send scope updates to that system, that system is always considered in scope
}
bool DataReplicator::IsObjectListPushComplete(PlayerID target)
{
	if (target==UNASSIGNED_PLAYER_ID)
		return false;

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(target);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return false;
	}

	return participantStruct->objectListPushComplete;
}
void DataReplicator::DereplicateObject(ReplicatedObject *object, bool sendNotificationPacket)
{
	DereplicateObjectInt(object, sendNotificationPacket, false);
}
void DataReplicator::DereplicateObjectInt(ReplicatedObject *object, bool sendNotificationPacket, bool remoteCall)
{
	assert(object);
	unsigned i;
	BaseObjectData *baseData = GetBaseObjectByReplicatedObject(object);
	if (baseData)
	{
		// If we own the object, or are the authority, then we do sends.  So delete the data we send to them and stop sending to each participant
		if (baseData->isOwner || baseData->replicatedObject->IsNetworkIDAuthority())
		{
			for (i=0; i < participantList.Size(); i++)
			{
				RemoveExtendedDataFromParticipant((BaseData *)baseData, participantList[i], false, sendNotificationPacket);
			}
		}
		objectList.RemoveElement(baseData->replicatedObject, BaseObjectDataComp);
		delete baseData;
	}

	// Go through the list of memory elements that are associated with this object and clear the associated object value
	if (object)
	{
		i=0;
		while (i < memoryList.Size())
		{
			if (memoryList[i]->associatedObject==object)
				UnsynchronizeMemoryInt(memoryList[i], sendNotificationPacket);
			else
				i++;
		}
	}

	if (remoteCall)
		object->OnRemoteDereplication(this);
}
void DataReplicator::RemoveExtendedDataFromParticipant(BaseData *baseData, ParticipantStruct *participantStruct, bool isMemory, bool sendNotificationPacket)
{
	bool objectExists;
	unsigned index;
	if (isMemory)
		index=participantStruct->extendedMemoryList.GetIndexFromKey(baseData, &objectExists, ExtendedMemoryDataComp);
	else
		index=participantStruct->extendedObjectList.GetIndexFromKey(baseData, &objectExists, ExtendedObjectDataComp);
	if (objectExists)
	{
		if (isMemory)
		{
			// If this is unified memory, we need to scan through all memory elements and count how many times this pointer is used.
			BaseMemoryData *baseMemoryData = (BaseMemoryData *) baseData;
			if (participantStruct->extendedMemoryList[index]->lastMemorySend)
			{
//#ifdef _DEBUG
//				assert(participantStruct->extendedMemoryList[index]->lastMemorySend->freeMemoryCB);
//#endif
				/*
				if (baseMemoryData->isUnifiedMemory)
				{
					unsigned count;
					count = GetUnifiedMemoryCount(participantStruct->extendedMemoryList[index]->lastMemorySend);
					if (count==1)
						baseMemoryData->freeMemoryCB(participantStruct->extendedMemoryList[index]->lastMemorySend, baseMemoryData->byteLength, baseMemoryData->context);
						//participantStruct->extendedMemoryList[index]->lastMemorySend->freeMemoryCB(participantStruct->extendedMemoryList[index]->lastMemorySend, participantStruct->extendedMemoryList[index]->lastMemorySend->context);
				}
				else
				*/
					baseMemoryData->freeMemoryCB(participantStruct->extendedMemoryList[index]->lastMemorySend, baseMemoryData->byteLength, baseMemoryData->context);
					//participantStruct->extendedMemoryList[index]->lastMemorySend->freeMemoryCB(participantStruct->extendedMemoryList[index]->lastMemorySend, participantStruct->extendedMemoryList[index]->lastMemorySend->context);
			}

			delete participantStruct->extendedMemoryList[index];
			participantStruct->extendedMemoryList.RemoveElementAtIndex(index);
		}
		else
		{
			if (sendNotificationPacket)
			{
				BaseObjectData *baseObjectData = (BaseObjectData *) baseData;
#ifdef _DEBUG
				assert(baseObjectData->isOwner || baseObjectData->replicatedObject->IsNetworkIDAuthority());
#endif
				SendObjectStopRequest(baseObjectData->replicatedObject->GetNetworkID(), participantStruct->playerId);
			}

			delete participantStruct->extendedObjectList[index];
			participantStruct->extendedObjectList.RemoveElementAtIndex(index);
		}
	}
}
unsigned DataReplicator::GetUnifiedMemoryCount(void *memory)
{
	unsigned i,j,count;
	for (i=0, count=0; i < participantList.Size(); i++)
	{
		for (j=0; j < (participantList[i])->extendedMemoryList.Size(); j++)
			if ((participantList[i])->extendedMemoryList[j]->lastMemorySend==memory)
				count++;
	}
	return count;
}
void* DataReplicator::GetSynchronizedMemoryCopy(BaseMemoryData *baseMemoryData, void *source)
{
//#ifdef _DEBUG
//	assert(baseMemoryData->synchronizedMemory->makeCopyCB);
//#endif
	/*
	if (baseMemoryData->isUnifiedMemory)
	{
		unsigned i,j;
		for (i=0; i < participantList.Size(); i++)
		{
			for (j=0; j < participantList[i]->extendedMemoryList.Size(); j++)
			{
				if (participantList[i]->extendedMemoryList[j]->baseData==baseMemoryData && participantList[i]->extendedMemoryList[j]->lastMemorySend)
				{
					return participantList[i]->extendedMemoryList[j]->lastMemorySend;
				}
			}
		}
		return baseMemoryData->makeCopyCB(source, baseMemoryData->byteLength, baseMemoryData->context);

		//return baseMemoryData->synchronizedMemory->makeCopyCB(baseMemoryData->synchronizedMemory, baseMemoryData->synchronizedMemory->context);
	}
	else
	*/
		return baseMemoryData->makeCopyCB(source, baseMemoryData->byteLength, baseMemoryData->context);
		//return baseMemoryData->synchronizedMemory->makeCopyCB(baseMemoryData->synchronizedMemory, baseMemoryData->synchronizedMemory->context);
}
void DataReplicator::Clear(void)
{
	unsigned i;

	while (memoryList.Size()>0)
		UnsynchronizeMemory(memoryList[memoryList.Size()-1]->synchronizedMemory, false);

	while (objectList.Size()>0)
	{
		if (objectList[objectList.Size()-1]->replicatedObject)
			DereplicateObjectInt(objectList[objectList.Size()-1]->replicatedObject, false, false);
	}

	for (i=0; i < participantList.Size(); i++)
	{
		delete participantList[i];
	}
	// Clear the participant list
	participantList.Clear();

	for (i=0; i < localObjectNameTable.size(); i++)
		delete localObjectNameTable[i];
	localObjectNameTable.clear();

	while (packetQueue.size())
		rakPeer->DeallocatePacket( packetQueue.pop() );
}
void DataReplicator::UnsynchronizeMemoryInt(BaseMemoryData *baseMemoryData, bool sendNotificationPacket)
{
	unsigned i;

	if (sendNotificationPacket)
	{
		for (i=0; i < participantList.Size(); i++)
			SendMemoryStopRequest(baseMemoryData->localIdentifier, participantList[i]->playerId);
	}

	for (i=0; i < participantList.Size(); i++)
	{
		unsigned index;
		for (index=0; index < participantList[i]->extendedMemoryList.Size(); index++)
			if (participantList[i]->extendedMemoryList[index]->baseData==baseMemoryData)
			{
				delete participantList[i]->extendedMemoryList[index];
				participantList[i]->extendedMemoryList.RemoveElementAtIndex(index);
				break;
			}
	}

	memoryList.RemoveElement(baseMemoryData->localKey, BaseMemoryDataComp);

	if (baseMemoryData->lastReceivedValue)
		baseMemoryData->freeMemoryCB(baseMemoryData->lastReceivedValue, baseMemoryData->byteLength, baseMemoryData->context);
	if (baseMemoryData->lastNonInterpValue)
		baseMemoryData->freeMemoryCB(baseMemoryData->lastNonInterpValue, baseMemoryData->byteLength, baseMemoryData->context);

	delete baseMemoryData;
}

void DataReplicator::ValidatedSend(ExtendedData *extendedData, ParticipantStruct *participant, bool isMemory, unsigned int currentTime, bool *objectPushDelayed)
{
	BaseData *baseData = extendedData->baseData;
	RakNet::BitStream outBitStream;
	bool validScope;
	ExtendedObjectData *extendedObjectData;
	BaseMemoryData *baseMemoryData;
	ExtendedMemoryData *extendedMemoryData;
	BaseObjectData *baseObjectData;

	if (baseData==0)
		return; // This data request doesn't have a corresponding synchronized element

	// Is this memory?
	if (isMemory)
	{
		bool dataDifferent;
		baseMemoryData = (BaseMemoryData*) baseData;
		extendedMemoryData = (ExtendedMemoryData*) extendedData;

		if (baseMemoryData->isOwner==false && (baseMemoryData->associatedObject==0 || baseMemoryData->associatedObject->IsNetworkIDAuthority()==false))
			return;

	//	if (extendedMemoryData->disableSendsUntilInterpolationComplete)
	//		return;

		//   Does this memory have an associated object ID?
		if (baseMemoryData->associatedObject)
		{
			// Yes, has associated object ID.  Can we find this object in the extended object send list for this participant?
			baseObjectData = GetBaseObjectByReplicatedObject(baseMemoryData->associatedObject);

			if (baseObjectData)
			{
				extendedObjectData = GetExtendedObjectByReplicatedObject(baseObjectData->replicatedObject, participant);
				if (baseObjectData->replicatedObject->GetWaitingForCreationResponse()==false)
				{
					if (extendedObjectData)
						validScope=extendedObjectData->objectInScope;
					else
						validScope=baseObjectData->replicatedObject->InScope(participant->playerId, this);
				}
				else
					validScope=false;
			}
			else
			{
				// Can't find the object.  Remove the base memory from the list
#ifdef _DEBUG
				assert(0);
#endif
				UnsynchronizeMemory(baseMemoryData->synchronizedMemory, false);
				return;
			}
		}
		else
		{
			//   No associated object.
			validScope=true;
		}

		if (validScope==false)
			return;

		bool includeTimestamp=false;
		RakNet::BitStream userData;
		void *serializeTarget;
		bool useInterpolatedValue=baseMemoryData->interpolateCB && baseMemoryData->isOwner==false && baseMemoryData->lastReceivedValue && baseMemoryData->needsInterpolation;

		//   Is the data different from what we last sent?
		//   if forcedComparisonDifference, then yes it is different
		//   Otherwise check to see if it is different.
		if (extendedMemoryData->lastMemorySend)
		{
			//if (baseMemoryData->isUnifiedMemory==false || baseMemoryData->didComparisonAndCopy==false)
			{
				// Since I'm not always sending synchronizedMemory, I need to compare the last sent value against the
				// appropriate value
				ShouldSendUpdateCallbackParameters ssucp;
				if (useInterpolatedValue)
					ssucp.memory=baseMemoryData->lastReceivedValue;
				else
					ssucp.memory=baseMemoryData->synchronizedMemory;
				ssucp.lastSentValue=extendedMemoryData->lastMemorySend;
				ssucp.memoryByteLength=baseMemoryData->byteLength;
				ssucp.destinationSystem=participant->playerId;
				ssucp.currentTime=currentTime;
				ssucp.lastSendTime=extendedMemoryData->lastSendTime;
				ssucp.context=baseMemoryData->context;
				
				dataDifferent=baseMemoryData->shouldSendUpdateCB(&ssucp);
			}
//			else
//				dataDifferent=baseMemoryData->dataDifferent;
		}
		else
		{
			// Don't send the constructor value if you are not the owner
			if (baseMemoryData->isOwner==false && baseMemoryData->gotFirstUpdate==false)
				return;

			dataDifferent=true;
		}

//		baseMemoryData->dataDifferent=dataDifferent;

		if (dataDifferent==false)
			return;

		// Encode new data value
//#ifdef _DEBUG
//		assert(baseMemoryData->synchronizedMemory->serializeCB);
//#endif

		// If we are interpolating and own the memory we are sending, we send the actual value
		// If we are not interpolating, we send the actual value
		// If we are interpolating and don't own the memory we are sending, we send the recieved value
		// If we don't support extrapolation, we send the last received value.
		if (useInterpolatedValue)
			serializeTarget=baseMemoryData->lastReceivedValue;
		else
			serializeTarget=baseMemoryData->synchronizedMemory;

		SerializeCallbackParameters scp;
		scp.sourceMemory=serializeTarget;
		scp.memoryByteLength=baseMemoryData->byteLength;
		scp.context=baseMemoryData->context;
		scp.outBitStream=&userData;
		if (extendedMemoryData->lastSendTime>0)
			scp.lastSentValue=extendedMemoryData->lastMemorySend;
		else
			scp.lastSentValue=0; // If never sent, pass 0 for lastMemorySend even if unified memory because you can't do delta sends for the first send to a player
		scp.currentTime=currentTime;
		scp.lastSendTime=extendedMemoryData->lastSendTime;
		scp.playerId=participant->playerId;
		scp.includeTimestamp=&includeTimestamp;

		baseMemoryData->serializeCB(&scp);

		if (extendedMemoryData->lastMemorySend==0)
			// First send so have to make or get a copy of the data
			extendedMemoryData->lastMemorySend = GetSynchronizedMemoryCopy(baseMemoryData, baseMemoryData->synchronizedMemory);

		// Encode timestamp, if requested
		if (includeTimestamp)
		{
			outBitStream.Write((unsigned char) ID_TIMESTAMP);
			if (useInterpolatedValue)
				outBitStream.Write(baseMemoryData->timePacketSent);
			else
				outBitStream.Write(RakNet::GetTime());
		}

		// Encode packet ID
		outBitStream.Write((unsigned char)ID_REPLICATOR_DATA_SEND_MEMORY);

		// Encode remote key
		if (extendedMemoryData->remoteKey==STR_TO_INDEX_MAP_MAX)
		{
			outBitStream.Write(false);
			stringCompressor->EncodeString(baseMemoryData->localIdentifier, 256, &outBitStream);
		}
		else
		{
			outBitStream.Write(true);
			outBitStream.Write(extendedMemoryData->remoteKey);
		}
		outBitStream.Write(&userData);

		// Send the packet
		rakPeer->Send(&outBitStream, baseMemoryData->priority, baseMemoryData->reliability, baseMemoryData->orderingChannel, participant->playerId, false);

//		if (baseMemoryData->isUnifiedMemory==false || baseMemoryData->didComparisonAndCopy==false)
//		{
			baseMemoryData->copyFromCB(extendedMemoryData->lastMemorySend, serializeTarget, baseMemoryData->byteLength, baseMemoryData->context);
//			baseMemoryData->didComparisonAndCopy=true;
//		}
		extendedMemoryData->lastSendTime=currentTime;
	}
	else
	{
		// (this is an object)
		baseObjectData = (BaseObjectData*) baseData;
		extendedObjectData = (ExtendedObjectData*) extendedData;

		if (baseObjectData->isOwner==false && baseObjectData->replicatedObject->IsNetworkIDAuthority()==false)
			return;

		if (baseObjectData->replicatedObject->GetWaitingForCreationResponse())
			return; // Don't use this object yet - waiting for a response from the server

		// Wrong: The owner of the object is not necessarily the requester.  We may need to send a push update
//		if (baseObjectData->replicatedObject->GetOwner()==participant->playerId)
//			return; // Don't bother sending scope or push updates to the owner of an object

		// Check to see if this object is currently in scope
		validScope=baseObjectData->replicatedObject->InScope(participant->playerId, this);

		// Is this object instantiated?
		if (extendedObjectData->objectInstantiated)
		{
			// Allow to send scope AND
			// Is the object in scope AND
			// Are we not sending the scope to the owner (owner of an object is by definition in scope relative to itself)
			if (baseObjectData->replicatedObject->IsNetworkIDAuthority() &&
				validScope!=extendedObjectData->objectInScope &&
				baseObjectData->replicatedObject->GetOwner()!=participant->playerId)
			{
				outBitStream.Write((unsigned char)ID_REPLICATOR_DATA_SEND_OBJECT_SCOPE);

				// Encode object ID
				outBitStream.Write(baseObjectData->replicatedObject->GetNetworkID());

				// Encode new scope
				outBitStream.Write(validScope);

				// Send the packet
				rakPeer->Send(&outBitStream, baseObjectData->priority, baseObjectData->reliability, baseObjectData->orderingChannel, participant->playerId, false);

				// Copy new scope to old scope.
				extendedObjectData->objectInScope=validScope;
				return;
			}
		}
		else
		{
			//  Is in scope?
			if (validScope && (baseObjectData->isOwner || baseObjectData->replicatedObject->IsNetworkIDAuthority()))
			{
				outBitStream.Write((unsigned char)ID_REPLICATOR_DATA_PUSH_OBJECT);

				// Encode objectID that the other system will use.
#ifdef _DEBUG
				assert(baseObjectData->replicatedObject->GetWaitingForCreationResponse()==false);
#endif
				outBitStream.Write(baseObjectData->replicatedObject->GetNetworkID());

				PlayerID objectOwner;
				objectOwner=baseObjectData->replicatedObject->GetOwner();
				if (objectOwner==rakPeer->GetInternalID())
					objectOwner=rakPeer->GetExternalID(participant->playerId);
				outBitStream.Write(objectOwner.binaryAddress);
				outBitStream.Write(objectOwner.port);

				// Encode remote key so the other system can look up the name of the object to create.  The remote
				// system doesn't yet have an assigned objectID.
				EncodeClassName(baseObjectData->replicatedObject->GetObjectClassName(), participant->playerId, &outBitStream);

				int createObject;
				createObject = baseObjectData->replicatedObject->OnSendReplicationPush(participant->playerId, &outBitStream, this);
                if (createObject==OBJECT_REPLICATION_OK)
				{
					// Send the packet
					rakPeer->Send(&outBitStream, baseObjectData->priority, RELIABLE_ORDERED,0, participant->playerId, false);

					// Record what we last sent by setting objectInScope and objectInstantiated to true
					extendedObjectData->objectInScope=true;
					extendedObjectData->objectInstantiated=true;
					return;
				}
				else if (createObject==OBJECT_REPLICATION_NEVER_PUSH)
				{
					// Never create the object for this player
					RemoveExtendedDataFromParticipant(baseData, participant, false, false);
				}
				// else
				{
					// try again later
					*objectPushDelayed=true;
				}
			}
		}
	}

}
/*
void DataReplicator::SendMemoryStartRequest(BaseMemoryData *baseData, PlayerID target)
{
	RakNet::BitStream outBitStream;

	// Encode the packet ID
	outBitStream.Write((unsigned char)ID_REPLICATOR_MEMORY_START_REQUEST);

	// Encode the local key
	outBitStream.Write(baseData->localKey);

	// Encode the string ID
	stringCompressor->EncodeString(baseData->localIdentifier, 256, &outBitStream);

	// Send the packet
	rakPeer->Send(&outBitStream, baseData->priority, RELIABLE_ORDERED, 0, target, false);
}
*/
void DataReplicator::SendMemoryStopRequest(const char *localIdentifier, PlayerID target)
{
	RakNet::BitStream outBitStream;

	// Encode the packet ID
	outBitStream.Write((unsigned char)ID_REPLICATOR_DATA_STOP);

	// Encode if this is memory or not
	outBitStream.Write(true);

	// Encode the local identifier (not the key, in case the other system didn't get the key yet).
	stringCompressor->EncodeString(localIdentifier, 256, &outBitStream);

	// Send the packet
	rakPeer->Send(&outBitStream, HIGH_PRIORITY, RELIABLE_ORDERED, 0, target, false);
}
void DataReplicator::SendObjectStopRequest(ObjectID networkID, PlayerID target)
{
	RakNet::BitStream outBitStream;

	// Encode the packet ID
	outBitStream.Write((unsigned char)ID_REPLICATOR_DATA_STOP);

	// Encode if this is memory or not
	outBitStream.Write(false);

	outBitStream.Write(networkID);

	// Send the packet
	rakPeer->Send(&outBitStream, HIGH_PRIORITY, RELIABLE_ORDERED, 0, target, false);
}
/*
void DataReplicator::SendDataStopRequest(BaseData *baseData, PlayerID target, bool isMemory)
{
	RakNet::BitStream outBitStream;

	// Encode the packet ID
	outBitStream.Write((unsigned char)ID_REPLICATOR_DATA_STOP);

	// Encode if this is memory or not
	outBitStream.Write(isMemory);

	if (isMemory)
	{
		BaseMemoryData *baseMemoryData = (BaseMemoryData *) baseData;

		// Encode the local identifier (not the key, in case the other system didn't get the key yet).
		stringCompressor->EncodeString(baseMemoryData->localIdentifier, 256, &outBitStream);
	}
	else
	{
		// Encode the object ID
		BaseObjectData *baseObjectData = (BaseObjectData *) baseData;
		outBitStream.Write(baseObjectData->replicatedObject->GetNetworkID());
	}

	// Send the packet
	rakPeer->Send(&outBitStream, baseData->priority, baseData->reliability==RELIABLE_ORDERED ? RELIABLE_ORDERED : RELIABLE, baseData->orderingChannel, target, false);
}
*/
DataReplicator::BaseMemoryData* DataReplicator::GetBaseMemoryDataByString(const char *str)
{
	unsigned i;
	for (i=0; i < memoryList.Size(); i++)
		if (strcmp(memoryList[i]->localIdentifier, str)==0)
			return memoryList[i];
	return 0;
}
//void DataReplicator::AddToParticipantList(PlayerID playerId, bool isLocalSystem)
void DataReplicator::AddToParticipantList(PlayerID playerId)
{
	// If this participant already exists, don't add because this call needs to be re-entrant
	if (GetParticipantByPlayerID(playerId))
		return;

	// If this participant is not connected, don't add them.  This reduces processing and fixes problems determining what to do for
	// participants with no data.
	if (rakPeer->GetIndexFromPlayerID(playerId)==-1)
		return;

	unsigned i;

	// Allocate a new ParticipantStruct
	ParticipantStruct *participantStruct;
	participantStruct = new ParticipantStruct;

	// Set the playerID of that ParticipantStruct
	participantStruct->playerId=playerId;

	participantStruct->sentObjectList=false;
	participantStruct->objectListPushComplete=false;

	//participantStruct->isLocalSystem=isLocalSystem;

	// Add the new ParticipantStruct to participantList
	participantList.InsertElement(participantStruct, participantStruct->playerId, ParticipantStructComp);

//	if (isLocalSystem==false)
//	{
		// For each object, if we send updates, add it to the list.

		for (i=0; i < objectList.Size(); i++)
		{
			if (objectList[i]->isOwner || objectList[i]->replicatedObject->IsNetworkIDAuthority())
			{
				// This will do an object push to that participant.  When that participant gets the push, they will request
				// that we send memory updates to them.
				AddExtendedObjectDataToParticipant(objectList[i], participantStruct);
			}
		}

		for (i=0; i < memoryList.Size(); i++)
			if (memoryList[i]->isOwner || (memoryList[i]->associatedObject && memoryList[i]->associatedObject->IsNetworkIDAuthority()))
				AddExtendedMemoryDataToParticipant(memoryList[i], GetParticipantByPlayerID(playerId));
//	}
}
void DataReplicator::RemoveFromParticipantList(PlayerID playerId, bool sendDataStopRequests, bool fromNetwork)
{
	ParticipantStruct *participantStruct;
	unsigned index;

	index = GetParticipantIndexByPlayerID(playerId);

	if (index==(unsigned)-1) // Yes, I know this is unsigned
	{
		return;
	}

	// Go through each object.  If the owner is playerId, call OnRemoteDereplication
	for (unsigned i=0; i < objectList.Size(); i++)
	{
		if (objectList[i]->replicatedObject->GetOwner()==playerId)
		{
			DereplicateObjectInt(objectList[i]->replicatedObject, sendDataStopRequests, fromNetwork);
		}
	}
	participantStruct=participantList[index];

	while (participantStruct->extendedMemoryList.Size())
		RemoveExtendedDataFromParticipant(participantStruct->extendedMemoryList[participantStruct->extendedMemoryList.Size()-1]->baseData, participantStruct, true, sendDataStopRequests);

	while (participantStruct->extendedObjectList.Size())
		RemoveExtendedDataFromParticipant(participantStruct->extendedObjectList[participantStruct->extendedObjectList.Size()-1]->baseData, participantStruct, false, sendDataStopRequests);


	delete participantStruct;
	participantList.RemoveElementAtIndex(index);
}

void DataReplicator::RemoveAllParticipants(void)
{
	while (participantList.Size())
		RemoveFromParticipantList(participantList[participantList.Size()-1]->playerId, false, false);
}
strToIndexMapType DataReplicator::GenerateMemoryStringMappingKey(void)
{
	strToIndexMapType newKey;

	// Increment the key as long as the key is in use.  Return the new key
	do {
		newKey=memoryToStringMappingKey++;
	} while(newKey==STR_TO_INDEX_MAP_MAX || GetMemoryDataByKey(newKey));

	return newKey;
}
DataReplicator::BaseMemoryData *DataReplicator::GetMemoryDataByKey(strToIndexMapType key)
{
	bool objectExists;
	unsigned index;

	index = memoryList.GetIndexFromKey(key, &objectExists, BaseMemoryDataComp);
	if (objectExists==false)
		return 0;

	return memoryList[index];
}
unsigned DataReplicator::GetParticipantIndexByPlayerID(PlayerID playerId)
{
	bool objectExists;
	unsigned index = participantList.GetIndexFromKey(playerId, &objectExists, ParticipantStructComp);
	if (objectExists==false)
		return (unsigned)-1;
	return index;
}
DataReplicator::ParticipantStruct *DataReplicator::GetParticipantByPlayerID(PlayerID playerId)
{
	unsigned index = GetParticipantIndexByPlayerID(playerId);
	if (index==(unsigned)-1)
		return 0;
	return participantList[index];
}
DataReplicator::ExtendedObjectData* DataReplicator::GetExtendedObjectByReplicatedObject(ReplicatedObject *object, ParticipantStruct *participant)
{
	unsigned index;
	ExtendedObjectData *extendedObjectData;
	BaseObjectData *baseObjectData;
#ifdef _DEBUG
	assert(participant);
#endif
	for (index=0; index < participant->extendedObjectList.Size(); index++)
	{
		extendedObjectData=participant->extendedObjectList[index];
		baseObjectData=(BaseObjectData *)extendedObjectData->baseData;
		if (baseObjectData->replicatedObject==object)
			return extendedObjectData;
	}
	return 0;
}
DataReplicator::ExtendedObjectData* DataReplicator::GetExtendedObjectByObjectID(ObjectID objectId, ParticipantStruct *participant)
{
	unsigned index;
	ExtendedObjectData *extendedObjectData;
	BaseObjectData *baseObjectData;
	for (index=0; index < participant->extendedObjectList.Size(); index++)
	{
		extendedObjectData=participant->extendedObjectList[index];
		baseObjectData=(BaseObjectData *)extendedObjectData->baseData;
		if (baseObjectData->replicatedObject->GetNetworkID()==objectId)
			return extendedObjectData;
	}
	return 0;
}
DataReplicator::ExtendedMemoryData* DataReplicator::GetExtendedMemoryByBaseMemory(BaseMemoryData *baseData, ParticipantStruct *participant)
{
	unsigned index;
	for (index=0; index < participant->extendedMemoryList.Size(); index++)
		if (participant->extendedMemoryList[index]->baseData==baseData)
			return participant->extendedMemoryList[index];
	return 0;
}
DataReplicator::BaseObjectData* DataReplicator::GetBaseObjectByReplicatedObject(ReplicatedObject *object)
{
	if (object==0)
		return 0;

	bool objectExists;
	unsigned index = objectList.GetIndexFromKey(object, &objectExists, BaseObjectDataComp);
	if (objectExists==false)
		return 0;
	return objectList[index];

}
DataReplicator::BaseObjectData* DataReplicator::GetBaseObjectByObjectID(ObjectID objectId)
{
	unsigned index;
	for (index=0; index < objectList.Size(); index++)
		if (objectList[index]->replicatedObject->GetNetworkID()==objectId)
			return objectList[index];
	return 0;
}
DataReplicator::BaseMemoryData *DataReplicator::GetBaseMemoryDataBySynchronizedMemory(const void *memory)
{
	unsigned i;
	for (i=0; i < memoryList.Size(); i++)
		if (memoryList[i]->synchronizedMemory==memory)
			return memoryList[i];
	return 0;
}

#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
void DataReplicator::OnDataOverflow(RakPeerInterface *peer, Packet *packet)
{
#ifdef _DEBUG
	assert(0);
#endif
}
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
void DataReplicator::OnDataUnderflow(RakPeerInterface *peer, Packet *packet)
{
#ifdef _DEBUG
	assert(0);
#endif
}
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
void DataReplicator::OnInvalidPacket(RakPeerInterface *peer, Packet *packet)
{
#ifdef _DEBUG
	assert(0);
#endif
}
void DataReplicator::OnAttach(RakPeerInterface *peer)
{
	rakPeer=peer;
}
#ifdef _MSC_VER
#pragma warning( disable : 4800 ) // warning C4100: <variable name> : unreferenced formal parameter
#endif
void DataReplicator::OnUpdate(RakPeerInterface *peer)
{
	unsigned i,j;
	ParticipantStruct *participantStruct;
	ExtendedData *extendedData;
	unsigned int currentTime=RakNet::GetTime();

	// We need to interpolate over time between the actual value and the extrapolated value, if we support extrapolation.
	// Otherwise we interpolate between the actual and last received value.
	BaseMemoryData *baseMemoryData;
	for (i=0; i < memoryList.Size(); i++)
	{
		baseMemoryData=memoryList[i];
		if (baseMemoryData->interpolateCB && baseMemoryData->needsInterpolation && baseMemoryData->userInterpolationFinished==false)
		{
			if (baseMemoryData->isOwner)
			{
				ShouldSendUpdateCallbackParameters ssucp;
				ssucp.memory=baseMemoryData->synchronizedMemory;
				ssucp.lastSentValue=baseMemoryData->lastNonInterpValue;
				ssucp.memoryByteLength=baseMemoryData->byteLength;
				ssucp.destinationSystem=UNASSIGNED_PLAYER_ID;
				ssucp.currentTime=currentTime;
				ssucp.lastSendTime=0;
				ssucp.context=0;

				// If the saved synchronized memory value is changed here, then it changed from the user.  Stop interpolating
				if (baseMemoryData->lastNonInterpValue && baseMemoryData->shouldSendUpdateCB(&ssucp))
				{
					// User changed the data and it wasn't from interpolateCB.  Stop interpolating.
					baseMemoryData->needsInterpolation=false;
				}
			}

			baseMemoryData->userInterpolationFinished=baseMemoryData->interpolateCB(baseMemoryData->synchronizedMemory, baseMemoryData->lastReceivedValue, currentTime, baseMemoryData->timePacketSent, baseMemoryData->lastInterpolateCallTime, baseMemoryData->context);
			baseMemoryData->lastInterpolateCallTime=currentTime;

			if (baseMemoryData->isOwner)
			{
				// Save synchronized memory value, so we know if it changed from the user
				if (baseMemoryData->lastNonInterpValue==0)
					baseMemoryData->lastNonInterpValue=baseMemoryData->makeCopyCB(baseMemoryData->synchronizedMemory, baseMemoryData->byteLength, baseMemoryData->context);
				else
					baseMemoryData->copyFromCB(baseMemoryData->lastNonInterpValue, baseMemoryData->synchronizedMemory, baseMemoryData->byteLength, baseMemoryData->context);
			}
		}
	}

	for (i=0; i < memoryList.Size(); i++)
	{
		memoryList[i]->doUpdate=false;
//		memoryList[i]->didComparisonAndCopy=false;
	}

	for (i=0; i < objectList.Size(); i++)
	{
		objectList[i]->doUpdate=false;

		if (objectList[i]->needsCreationRequest)
		{
			RakNet::BitStream outContext;
			BaseObjectData *baseObjectData;
			baseObjectData=objectList[i];
			ObjectReplicationPushResult pushResult = baseObjectData->replicatedObject->OnSendCreationRequest(baseObjectData->replicatedObject->GetNetworkIDAuthority(), &outContext, this);
			if (pushResult==OBJECT_REPLICATION_OK)
			{
				RequestObjectCreation(0, &outContext, baseObjectData->replicatedObject, baseObjectData->priority, baseObjectData->reliability, baseObjectData->orderingChannel, baseObjectData->replicatedObject->GetNetworkIDAuthority());
				objectList[i]->needsCreationRequest=false;
			}
			else if (pushResult==OBJECT_REPLICATION_NEVER_PUSH)
			{
				objectList[i]->needsCreationRequest=false;
			}
		}
	}

	if (participantList.Size()>0)
	{
		// Keep track if any object pushes were delayed.  If so, we don't send ID_REPLICATOR_PUSH_COMPLETE.  This way, that packet is only
		// sent if either all the objects were either pushed or never pushed.
		bool objectPushDelayed=false;

		for (i=0; i < participantList.Size(); i++)
		{
			participantStruct=participantList[i];

			for (j=0; j < participantStruct->extendedObjectList.Size(); j++)
			{
				extendedData=participantStruct->extendedObjectList[j];

				if ( extendedData->baseData->doUpdate || extendedData->baseData->lastUpdateTime + extendedData->baseData->maxUpdateFrequencyMS < currentTime)
				{
					// This will send object scope updates and object pushes
					extendedData->baseData->doUpdate=true;
					ValidatedSend(extendedData, participantStruct, false, currentTime, &objectPushDelayed);
					extendedData->baseData->lastUpdateTime=currentTime;
				}
			}

			if (participantStruct->sentObjectList==false && objectPushDelayed==false)
			{
				participantStruct->sentObjectList=true;

				// Tell this system the first batch of push calls went through
				unsigned char c=ID_REPLICATOR_PUSH_COMPLETE;
				rakPeer->Send((const char*) &c, sizeof(unsigned char), LOW_PRIORITY, RELIABLE_ORDERED,0, participantStruct->playerId, false);                
			}
		}
	}

	for (i=0; i < objectList.Size(); i++)
	{
		if (objectList[i]->calledOnReplication==false && objectList[i]->replicatedObject->GetWaitingForCreationResponse()==false)
		{
			// User will synch memory here
			objectList[i]->calledOnReplication=true;
			objectList[i]->replicatedObject->OnReplication(this);
		}
	}

	if (participantList.Size()>0)
	{
		for (i=0; i < participantList.Size(); i++)
		{
			participantStruct=participantList[i];

			for (j=0; j < participantStruct->extendedMemoryList.Size(); j++)
			{
				extendedData=participantStruct->extendedMemoryList[j];

				if ( extendedData->baseData->doUpdate || extendedData->baseData->lastUpdateTime + extendedData->baseData->maxUpdateFrequencyMS < currentTime)
				{
					// This will send synched memory
					extendedData->baseData->doUpdate=true;
					ValidatedSend(extendedData, participantStruct, true, currentTime, 0);
					extendedData->baseData->lastUpdateTime=currentTime;
				}
			}
		}
	}
}
// We got a memory update
void DataReplicator::OnDataReplicateSendMemory(RakPeerInterface *peer, Packet *packet)
{
	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);
	unsigned char typeId;

	incomingBitStream.Read(typeId);

	unsigned int timePacketSent;
	if (typeId==ID_TIMESTAMP)
	{
		incomingBitStream.Read(timePacketSent);
		incomingBitStream.IgnoreBits(8);
	}
	else
		timePacketSent=0;

	strToIndexMapType transmittedKey;
	bool hasTransmittedKey;
	unsigned i;
	bool sendStrMapIndex;
	strToIndexMapType mapIndex;
	sendStrMapIndex=false;

	char identifier[256];
	identifier[0]=0;
	incomingBitStream.Read(hasTransmittedKey);
	if (hasTransmittedKey)
	{
		if (!incomingBitStream.Read(transmittedKey))
		{
			OnDataUnderflow(peer, packet);
			return;
		}
	}
	else
	{
		if (!stringCompressor->DecodeString(identifier, 256, &incomingBitStream))
		{
			OnDataUnderflow(peer, packet);
			return;
		}

		transmittedKey=STR_TO_INDEX_MAP_MAX;
		for (i=0; i < memoryList.Size(); i++)
		{
			if (strcmp(memoryList[i]->localIdentifier, identifier)==0)
			{
				transmittedKey=memoryList[i]->localKey;
				break;
			}
		}

		if (transmittedKey==STR_TO_INDEX_MAP_MAX)
			SendMemoryStopRequest(identifier, packet->playerId);
		else
		{
			sendStrMapIndex=true;
			mapIndex=memoryList[i]->localKey;
		}
	}

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	BaseMemoryData *baseMemoryData = (BaseMemoryData *)GetMemoryDataByKey(transmittedKey);
	void *source;
	// Allow memory updates if we are not the owner and the owner is sending to us, or if we are a client
	if (baseMemoryData &&
		((baseMemoryData->isOwner==false && (baseMemoryData->associatedObject==0 || baseMemoryData->associatedObject->GetOwner()==packet->playerId)) ||
		(baseMemoryData->associatedObject && baseMemoryData->associatedObject->IsNetworkIDAuthority()==false)))
	{
		baseMemoryData->gotFirstUpdate=true;

		if (sendStrMapIndex)
		{
			RakNet::BitStream mapping;
			mapping.Write((unsigned char)ID_REPLICATOR_STR_MAP_INDEX);
			mapping.Write(false); // memory
			mapping.Write(mapIndex);
			stringCompressor->EncodeString(identifier, 256, &mapping);
			rakPeer->Send(&mapping, HIGH_PRIORITY, UNRELIABLE, 0, packet->playerId, false);
		}

		unsigned currentTime = RakNet::GetTime();

		if (timePacketSent)
			baseMemoryData->timePacketSent=timePacketSent;
		else
			baseMemoryData->timePacketSent=currentTime;

		DeserializeCallbackParameters dcp;
		dcp.destinationMemory=baseMemoryData->lastReceivedValue;
		dcp.memoryByteLength=baseMemoryData->byteLength;
		dcp.context=baseMemoryData->context;
		dcp.inBitStream=&incomingBitStream;
		dcp.currentTime=currentTime;
		dcp.timePacketSent=timePacketSent;
		dcp.lastTimeCalled=baseMemoryData->lastReceivedTime;
		dcp.playerId=packet->playerId;

		// If we receive a memory block and allow interpolation, we need to store a copy of the last received value.
		if (baseMemoryData->interpolateCB)
		{
			baseMemoryData->lastInterpolateCallTime = baseMemoryData->timePacketSent;
			baseMemoryData->needsInterpolation=true;
			baseMemoryData->userInterpolationFinished=false;

			// Allocate the last lastReceivedValue if we do interpolation
			if (baseMemoryData->lastReceivedValue==0)
				baseMemoryData->lastReceivedValue=baseMemoryData->makeCopyCB(baseMemoryData->synchronizedMemory, baseMemoryData->byteLength, baseMemoryData->context);

			// If we allow interpolation, write updates to the last received value, rather than the actual value.
			baseMemoryData->deserializeCB(&dcp);
			baseMemoryData->lastReceivedTime=currentTime;
			source=baseMemoryData->lastReceivedValue;
		}
		else
		{
			baseMemoryData->deserializeCB(&dcp);
			baseMemoryData->lastReceivedTime=currentTime;
			source=baseMemoryData->synchronizedMemory;
		}

		if (baseMemoryData->isOwner || (baseMemoryData->associatedObject && baseMemoryData->associatedObject->IsNetworkIDAuthority()))
		{
			// Don't send back to the sender
			ExtendedMemoryData *extendedMemoryData = GetExtendedMemoryByBaseMemory(baseMemoryData, participantStruct);
			if (extendedMemoryData)
			{
				if (extendedMemoryData->lastMemorySend==0)
					extendedMemoryData->lastMemorySend = GetSynchronizedMemoryCopy(baseMemoryData, source);
				else
					baseMemoryData->copyFromCB(extendedMemoryData->lastMemorySend, source, baseMemoryData->byteLength, baseMemoryData->context);
			}
		}
	}
	else
	{

#ifdef _DEBUG
		// This only matters in debug for my own testing - we got a memory update for unknown or unwritable memory.  This happens normally in the regular situation
	//	assert(0);
		printf("Warning: Got a memory update for unknown or unwritable memory\n");
		SendMemoryStopRequest(identifier, packet->playerId);
#endif
	}

}
// We got a packet telling us to create an object and that that object is in scope
void DataReplicator::OnSendReplicationPushPacket(RakPeerInterface *peer, Packet *packet)
{
	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);
	incomingBitStream.IgnoreBits(8);

	ObjectID objectId;
	PlayerID objectOwner;

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

    incomingBitStream.Read(objectId);
	if (objectId==UNASSIGNED_OBJECT_ID)
	{
		OnInvalidPacket(peer, packet);
		return;
	}

	if (NetworkIDGenerator::GET_BASE_OBJECT_FROM_ID(objectId))
		return; // Object already exists

	incomingBitStream.Read(objectOwner.binaryAddress);
	incomingBitStream.Read(objectOwner.port);

    char className[256];
	if (DecodeClassName(className, participantStruct, &incomingBitStream, peer, packet)==false)
	{
#ifdef _DEBUG
		// Only matters in debug for an unknown string.  In production this could happen anytime from hackers or cheaters.
		assert(0);
#endif
		return;
	}

	ReplicatedObject *newObject = CreateObject(className, &incomingBitStream);
	if (newObject==0)
		return;

	BaseObjectData *baseObjectData;
	bool res;

	res=newObject->OnReceiveReplicationPush(packet->playerId, &incomingBitStream, objectId, objectOwner, this);
	if (res==false)
	{
		DereplicateObject(newObject, false);
		delete newObject;
		newObject=0;
	}

	if (objectOwner==rakPeer->GetExternalID(packet->playerId))
		objectOwner=rakPeer->GetInternalID();
	// If the user registered the object, we may need to change the property flags if it was autoassigned.
	// We also need to re-add the extended object data, in case the owner was changed
	baseObjectData=GetBaseObjectByReplicatedObject(newObject);
	/*
	if (newObject && baseObjectData)
	{
		if (baseObjectData->autoAssignPropertyFlags)
			AutoAssignPropertyFlags(baseObjectData, newObject, (objectOwner==rakPeer->GetInternalID())!=0);
	}
	*/

	OnCreationCommand(newObject, packet->playerId, objectId, objectOwner, baseObjectData);
}

// Packet telling us an object's scope has changed
void DataReplicator::OnDataReplicateSendObjectScope(RakPeerInterface *peer, Packet *packet)
{
	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);

	// SetWriteOffset is used here to get around a design flaw, where I should have had the bitstream constructor take bits, rather than bytes
	// It sets the actual number of bits in the packet
	incomingBitStream.SetWriteOffset(packet->bitSize);
	incomingBitStream.IgnoreBits(8);

	ObjectID objectId;
	bool newScope;

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	incomingBitStream.Read(objectId);
	if (!incomingBitStream.Read(newScope))
	{
		OnDataUnderflow(peer, packet);
		return;
	}

	BaseObjectData *baseObjectData = GetBaseObjectByObjectID(objectId);

	if (baseObjectData)
	{
		if (baseObjectData->replicatedObject->IsNetworkIDAuthority()==false)
		{
			baseObjectData->replicatedObject->OnScopeChange(packet->playerId, newScope, this);

			if (incomingBitStream.GetNumberOfUnreadBits()!=0)
				OnDataOverflow(peer,packet);
		}
		else
		{
#ifdef _DEBUG
			// We set not to receive scope updates but got one
			assert(0);
#endif
		}
	}
	else
	{
#ifdef _DEBUG
		// This only matters in debug for my own testing
		//assert(0);
		printf("Warning: Got scope update for unknown object\n");
#endif
	}
}
/*
// Packet telling us to start sending data.  This will add an extended element to
// the participant list assuming that extended element is set to send.
// It also sets up a string to remote key mapping
void DataReplicator::OnMemoryReplicateStart(RakPeerInterface *peer, Packet *packet)
{
	char identifier[256];
	unsigned short transmittedKey;

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	identifier[0]=0;

	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);
	incomingBitStream.SetWriteOffset(packet->bitSize);
	incomingBitStream.IgnoreBits(8);

	incomingBitStream.Read(transmittedKey);

	if (!stringCompressor->DecodeString(identifier, 256, &incomingBitStream))
	{
		OnDataUnderflow(peer, packet);
		return;
	}

	// Make sure the base data does not already have this player set to send data
	BaseMemoryData *baseData = GetBaseMemoryDataByString(identifier);
	if (baseData==0)
	{
		// We don't have memory registered with this string
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	if (baseData->isOwner==false && (baseData->associatedObject && baseData->associatedObject->IsNetworkIDAuthority()==false))
	{
		// Asked to start sending updates when this is receive only
		// Shouldn't assert - the other system doesn't know we are receive only.  The other system only knows that it receives
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	// If this participant already has this key, this is a duplicate packet
	if (participantStruct->GetExtendedMemoryIndexByKey(transmittedKey)!=(unsigned)-1)
	{
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	if (participantStruct->extendedMemoryList.HasData(baseData, ExtendedMemoryDataComp))
		return;

	ExtendedMemoryData *extendedData = new ExtendedMemoryData;
	extendedData->remoteKey=transmittedKey;
	extendedData->baseData=baseData;
	extendedData->lastMemorySend=0;
	extendedData->lastSendTime=0;

	participantStruct->extendedMemoryList.InsertElement(extendedData, extendedData->baseData, ExtendedMemoryDataComp);

	if (incomingBitStream.GetNumberOfUnreadBits()!=0)
		OnDataOverflow(peer, packet);
}
*/
// Packet telling us to stop sending data.  This will remove an extended element in
// the participant list assuming that extended element is set to send
void DataReplicator::OnDataReplicateStop(RakPeerInterface *peer, Packet *packet)
{
	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);
	incomingBitStream.SetWriteOffset(packet->bitSize);
	incomingBitStream.IgnoreBits(8);

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	bool isMemory;
	BaseData *baseData;
	incomingBitStream.Read(isMemory);
	if (isMemory)
	{
		char str[256];
		if (!stringCompressor->DecodeString(str, 256, &incomingBitStream))
		{
			OnDataUnderflow(peer, packet);
			return;
		}
		baseData = GetBaseMemoryDataByString(str);
	}
	else
	{
		ObjectID objectId;
		if (!incomingBitStream.Read(objectId))
		{
			OnDataUnderflow(peer, packet);
			return;
		}
		baseData = GetBaseObjectByObjectID(objectId);
	}

	if (baseData==0)
	{
		// Don't know what they are referring to
		return;
	}

	if (isMemory)
	{
		BaseMemoryData *baseMemoryData = (BaseMemoryData*)baseData;
		if (baseMemoryData->isOwner || (baseMemoryData->associatedObject && baseMemoryData->associatedObject->IsNetworkIDAuthority()))
			RemoveExtendedDataFromParticipant(baseData, participantStruct, true, false);
		else
			UnsynchronizeMemory(((BaseMemoryData*)baseData)->synchronizedMemory, false);
	}
	else
	{
		ReplicatedObject *object = ((BaseObjectData*)baseData)->replicatedObject;
		BaseObjectData *baseObjectData = (BaseObjectData*)baseData;
		bool unreplicateObject = object->AcceptDereplicationRequest(packet->playerId, this);

		if (baseObjectData->isOwner || baseObjectData->replicatedObject->IsNetworkIDAuthority())
		{
			if (unreplicateObject)
			{
				RemoveExtendedDataFromParticipant(baseData, participantStruct, false, false);
				DereplicateObjectInt(object, true, true);
			}
			else
			{
				// Remove the object for only the sender.
				RemoveExtendedDataFromParticipant(baseData, participantStruct, false, false);
			}
		}

		if (baseObjectData->isOwner==false)
		{
			if (unreplicateObject)
			{
				// Remove the object for ourselves
				DereplicateObjectInt(object, false, true);
			}
		}
	}
}
void DataReplicator::OnNewPlayer(RakPeerInterface *peer, Packet *packet)
{
	// AddParticipant for this player if broadcast to all.
	// Otherwise it's up to the user to add them.
	if (broadcastToAll)
		AddParticipant(packet->playerId);
}

void DataReplicator::OnLostPlayer(RakPeerInterface *peer, Packet *packet)
{
	RemoveFromParticipantList(packet->playerId, true, true);
}
bool DataReplicator::SendImmediateMemoryStartRequest(BaseMemoryData *memoryData, ParticipantStruct *participantStruct)
{
	// If a global variable, just send it.
	if (memoryData->associatedObject==0)
		return true;

	/*
	// If we own own the object the memory is associated with, send the start request.  This means that the object has already been instantiated.
	BaseObjectData *baseObjectData;
	baseObjectData = GetBaseObjectByReplicatedObject(memoryData->associatedObject);
	if (baseObjectData->isOwner)
		return true;
		*/

	// Otherwise, we don't own the object.  Only send the memory start request if it has been instantiated on the other system.
	ExtendedObjectData *extendedObjectData;
	extendedObjectData = GetExtendedObjectByReplicatedObject(memoryData->associatedObject, participantStruct);
	return extendedObjectData && extendedObjectData->objectInstantiated;
}
void DataReplicator::EncodeClassName(const char *strIn, PlayerID playerId, RakNet::BitStream *bitStream)
{
	ParticipantStruct *participantStruct=GetParticipantByPlayerID(playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	unsigned i;
	for (i=0; i < participantStruct->remoteObjectNameTable.size(); i++)
	{
		if (participantStruct->remoteObjectNameTable[i] && strcmp(strIn, participantStruct->remoteObjectNameTable[i])==0)
		{
			bitStream->Write(false);
			bitStream->Write((strToIndexMapType) i);
			return;
		}
	}

	bitStream->Write(true);
	stringCompressor->EncodeString((const char*)strIn, 256, bitStream);
}
void DataReplicator::OnStringMapIndex(RakPeerInterface *peer, Packet *packet)
{
	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);
	incomingBitStream.SetWriteOffset(packet->bitSize);
	incomingBitStream.IgnoreBits(8);

	strToIndexMapType index;
	char str[256];
	bool isObject;

	incomingBitStream.Read(isObject);

	incomingBitStream.Read(index);
	if (!stringCompressor->DecodeString(str, 256, &incomingBitStream))
	{
		OnDataUnderflow(peer, packet);
		return;
	}

	if (index==STR_TO_INDEX_MAP_MAX)
	{
		OnInvalidPacket(peer, packet);
		return;
	}

	if (incomingBitStream.GetNumberOfUnreadBits()!=0)
	{
		OnDataOverflow(peer,packet);
		return;
	}

	if (str[0]==0)
	{
		OnInvalidPacket(peer, packet);
		return;
	}

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	if (isObject)
	{
		if (index < participantStruct->remoteObjectNameTable.size() && participantStruct->remoteObjectNameTable[index]!=0)
		{
			// Make sure the same string is sent to the same slot
#ifdef _DEBUG
			assert(strcmp(participantStruct->remoteObjectNameTable[index], str)==0);
#endif
			return;
		}


		char *allocatedStr = new char [strlen(str)+1];
		strcpy(allocatedStr, str);
		if (isObject)
		{
			if (participantStruct->remoteObjectNameTable.size() < index)
				delete [] participantStruct->remoteObjectNameTable[index];
			participantStruct->remoteObjectNameTable.replace(allocatedStr,0,index);
		}
	}
	else
	{
		BaseMemoryData *baseMemoryData = GetBaseMemoryDataByString(str);
		if (baseMemoryData)
		{
			ExtendedMemoryData *extendedMemoryData = GetExtendedMemoryByBaseMemory(baseMemoryData, participantStruct);
			if (extendedMemoryData)
			{
				extendedMemoryData->remoteKey=index;
			}
		}
	}
}
//ObjectReplicationHandler*
bool DataReplicator::DecodeClassName(char *strOut, ParticipantStruct *participantStruct, RakNet::BitStream *bitStream, RakPeerInterface *peer, Packet *packet)
{
	bool hasEncodedString;
	strToIndexMapType indexToSend=STR_TO_INDEX_MAP_MAX;
	bitStream->Read(hasEncodedString);
	if (hasEncodedString)
	{
		if (!stringCompressor->DecodeString(strOut, 256, bitStream))
		{
			OnDataUnderflow(peer, packet);
			return false;
		}

		if (strOut[0]==0)
			return false;

		if (strlen(strOut)>=256)
			return false;

		unsigned i;
		bool alreadyHasString=false;
		for (i=0; i < localObjectNameTable.size(); i++)
		{
			if (strcmp(localObjectNameTable[i], strOut)==0)
			{
				alreadyHasString=true;
				break;
			}
		}

        if (alreadyHasString==false)
		{
			// Add a string to the table and send the index
			if (localObjectNameTable.size()+1<STR_TO_INDEX_MAP_MAX)
			{
				char *strCpy;
				strCpy = new char [strlen(strOut)+1];
				strcpy(strCpy, strOut);
				localObjectNameTable.insert(strCpy);

				// Send the index localObjectNameTable.size() with the string
				indexToSend=(strToIndexMapType)(localObjectNameTable.size()-1);
			}
			else
			{
#ifdef _DEBUG
				// strToIndexMapType is too small
				assert(0);
#endif
			}

		}
		else
		{
			// Send the index i with the string
			indexToSend=(strToIndexMapType)i;
		}
	}
	else
	{
		strToIndexMapType indexMap;
		bitStream->Read(indexMap);
		if (indexMap >= localObjectNameTable.size() )
		{
			// Invalid index
#ifdef _DEBUG
			assert(0);
#endif
			return false;
		}

		if (localObjectNameTable[indexMap]==0)
		{
			// Unknown string
#ifdef _DEBUG
			assert(0);
#endif
			return false;
		}

		strcpy(strOut, localObjectNameTable[indexMap]);
	}

	if (indexToSend!=STR_TO_INDEX_MAP_MAX)
	{
		RakNet::BitStream outBitStream;
		outBitStream.Write((unsigned char)ID_REPLICATOR_STR_MAP_INDEX);
		outBitStream.Write(true); // object
		outBitStream.Write(indexToSend);
		stringCompressor->EncodeString(strOut, 256, &outBitStream);
		rakPeer->Send(&outBitStream, HIGH_PRIORITY, UNRELIABLE, 0, participantStruct->playerId, false);
	}

	//return objectReplicationHandler;
	return true;
}
void DataReplicator::OnObjectCreationRequest(RakPeerInterface *peer, Packet *packet)
{
	unsigned char creationRequestIndex;
	char className[256];
	RakNet::BitStream outBitStream, context;
	ReplicatedObject* newObject;
	ExtendedObjectData *extendedObjectData;

	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);
	incomingBitStream.SetWriteOffset(packet->bitSize);
	incomingBitStream.IgnoreBits(8);

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	//ObjectReplicationHandler* objectReplicationHandler =
	if (DecodeClassName(className, participantStruct, &incomingBitStream, peer, packet)==false)
	{
		// Unknown request.  Only something to be concerned about in development.  In release this could happen from cheaters or hackers.
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}


	//if (hasCreationRequestIndex)
	//{
		if (!incomingBitStream.Read(creationRequestIndex))
		{
			OnDataUnderflow(peer, packet);
			return;
		}
	//}


	// Read owner
	PlayerID objectOwner;
	incomingBitStream.Read(objectOwner.binaryAddress);
	incomingBitStream.Read(objectOwner.port);

	newObject = CreateObject(className, &incomingBitStream);
	if (newObject==0)
		return;
	assert(newObject->IsNetworkIDAuthority());
	if (newObject->GetOwner()!=UNASSIGNED_PLAYER_ID)
		objectOwner=newObject->GetOwner(); // The object constructor could have set the owner by the ReplicateObject call

	bool allowNewObject = newObject->OnReceiveCreationRequest(packet->playerId, &incomingBitStream, &context, objectOwner, this);
	if (allowNewObject==false)
	{
		delete newObject;
		return;
	}

	assert(objectOwner!=UNASSIGNED_PLAYER_ID); // Object owner must be set either by the requester, by a ReplicateObject call from the constructor, or by the user in OnReceiveCreationRequest
	newObject->SetOwner(objectOwner);

	// If the object was already registered and the owner was changed, be sure to update the property flags to match it
	BaseObjectData *baseObject;
	baseObject=GetBaseObjectByReplicatedObject(newObject);
	if (objectOwner==rakPeer->GetExternalID(packet->playerId))
		objectOwner=rakPeer->GetInternalID();

	//newObject = objectReplicationHandler->OnReceiveCreationRequest(packet->playerId, className, &incomingBitStream, &context, &priority, &reliability, &orderingChannel, &maxUpdateFrequencyMS);
	outBitStream.Write((unsigned char)ID_REPLICATOR_OBJECT_CREATION_REQUEST_RESPONSE);
//	outBitStream.Write(hasCreationRequestIndex);
//	if (hasCreationRequestIndex)
		outBitStream.Write(creationRequestIndex);

	outBitStream.Write(allowNewObject);

	EncodeClassName(className, packet->playerId, &outBitStream);

	if (allowNewObject)
	{
#ifdef _DEBUG
		// Not a good idea to instantiate objects that are not in scope to the requester or it will just immediately go out of scope.
		// By definition instantiated objects are in-scope
		assert(newObject->InScope(packet->playerId, this));
#endif

//		assert(GetExtendedObjectByObjectID(newObject->GetNetworkID());

		if (objectOwner==rakPeer->GetInternalID())
			objectOwner=rakPeer->GetExternalID(packet->playerId);

		// Write objectOwner in packet
		outBitStream.Write(objectOwner.binaryAddress);
		outBitStream.Write(objectOwner.port);

		outBitStream.Write(newObject->GetNetworkID());
	}

	outBitStream.Write(&context);

	rakPeer->Send(&outBitStream, HIGH_PRIORITY, RELIABLE_ORDERED, 0, packet->playerId, false);

	// Send the memory start after the object creation request is responded to
	if (allowNewObject)
	{
		// packet->playerId, instead of owner, is intentional.  These operations are for the requester of an object, not for the owner.
		extendedObjectData = GetExtendedObjectByObjectID(newObject->GetNetworkID(), GetParticipantByPlayerID(packet->playerId));
		if (extendedObjectData)
		{
			// Don't push to the sender
			extendedObjectData->objectInScope=true;
			extendedObjectData->objectInstantiated=true;
		}
	}
}
void DataReplicator::OnPushComplete(RakPeerInterface *peer, Packet *packet)
{
	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

#ifdef _DEBUG
	assert(participantStruct->objectListPushComplete==false);
#endif

	participantStruct->objectListPushComplete=true;
}
void DataReplicator::OnObjectCreationRequestResponse(RakPeerInterface *peer, Packet *packet)
{
	RakNet::BitStream incomingBitStream((char*)packet->data, packet->length, false);
	incomingBitStream.SetWriteOffset(packet->bitSize);
	incomingBitStream.IgnoreBits(8);
	bool requestAccepted;
	ReplicatedObject* newObject;
	ObjectID objectId;
	unsigned char creationRequestIndex;
	ReplicatedObject *object=0;

	ParticipantStruct *participantStruct=GetParticipantByPlayerID(packet->playerId);
	if (participantStruct==0)
	{
		// Got an update from someone that is not a participant
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	// Record requests so we don't accept responses when we never sent a request to begin with.
	// Record request by player rather than by object so the player can change what kind of object is created
	if (participantStruct->objectRequestCount==0)
	{
		// Got more object request responses than we sent requests for!
		OnInvalidPacket(peer, packet);
		return;
	}

	participantStruct->objectRequestCount--;

//	if (!incomingBitStream.Read(hasCreationRequestIndex))
//	{
//		OnDataUnderflow(peer, packet);
//		return;
//	}
//	if (hasCreationRequestIndex)
//	{
		if (!incomingBitStream.Read(creationRequestIndex))
		{
			OnDataUnderflow(peer, packet);
			return;
		}
		else
		{
			if (creationRequestIndex < LOCAL_CREATION_REQUEST_TABLE_SIZE && localCreationRequestTable[creationRequestIndex])
			{
				object=localCreationRequestTable[creationRequestIndex];
				localCreationRequestTable[creationRequestIndex]=0;
#ifdef _DEBUG
				assert(GetBaseObjectByReplicatedObject(object) && GetBaseObjectByReplicatedObject(object)->replicatedObject->GetNetworkIDAuthority()==packet->playerId);
#endif
			}
			else
			{
				OnInvalidPacket(peer, packet);
				return;
			}
		}
//	}

	if (!incomingBitStream.Read(requestAccepted))
	{
		OnDataUnderflow(peer, packet);
		return;
	}

	char className[256];
	if (DecodeClassName(className, participantStruct, &incomingBitStream, peer, packet)==false)
	{
		// Unknown request.  Only something to be concerned about in development.  In release this could happen from cheaters or hackers.
#ifdef _DEBUG
		assert(0);
#endif
		return;
	}

	if (requestAccepted)
	{
		// read object owner from packet
		PlayerID objectOwner;
		incomingBitStream.Read(objectOwner.binaryAddress);
		incomingBitStream.Read(objectOwner.port);

		if (!incomingBitStream.Read(objectId))
		{
			OnDataUnderflow(peer, packet);
			return;
		}

		if (object)
		{
			newObject = object;
		}
		else
		{
			newObject = CreateObject(className, &incomingBitStream);
		}

		if (!newObject->OnCreationRequestAcceptance(packet->playerId, objectId, &incomingBitStream, objectOwner, this))
		{
			if (objectId==UNASSIGNED_OBJECT_ID)
				delete newObject;
			newObject=0;
		}

		// If the object was already registered and the owner was changed, be sure to update the property flags to match it
		BaseObjectData *baseObject;
		baseObject=GetBaseObjectByReplicatedObject(newObject);
		/*
		if (baseObject && baseObject->autoAssignPropertyFlags)
		{
			AutoAssignPropertyFlags(baseObject, newObject, (objectOwner==rakPeer->GetExternalID(packet->playerId))!=0);
		}
		*/

//		ReplicatedObject* newObject = objectReplicationHandler->OnCreationRequestAcceptance(packet->playerId, object, className, objectId, &incomingBitStream);
		OnCreationCommand(newObject, packet->playerId, objectId, objectOwner, baseObject);
	}
	else
	{
		newObject = CreateObject(className, &incomingBitStream);
		if (newObject==0)
			return;
		newObject->OnCreationRequestRejection(packet->playerId, &incomingBitStream, this);
		delete newObject;
	}
}

ReplicatedObject *DataReplicator::CreateObject(const char *className, RakNet::BitStream *inContext)
{
	// create the class - do not replicate again.
	assert(objectFactoryCallback);

	// This is so the constructor of the object doesn't replicate the object, so we don't have the network creating an object, then the object
	// replicating itself, which in turns creates the object again or sends another replicate request
	ReplicatedObject *objectToReplicate = objectFactoryCallback(className, inContext);
	return objectToReplicate;
}
/*
void DataReplicator::AutoAssignPropertyFlags(BaseObjectData *baseObjectData, ReplicatedObject *newObject, bool isObjectOwner)
{
#ifdef _DEBUG
	assert(baseObjectData->autoAssignPropertyFlags);
#endif
	if (rakPeer->GetInternalID()==newObject->GetNetworkIDAuthority())
	{
		if (isObjectOwner)
			baseObjectData->propertyFlags=OBJECT_FLAG_SERVER_OWNER;
		else
			baseObjectData->propertyFlags=OBJECT_FLAG_SERVER_NON_OWNER;
	}
	else
	{
		if (isObjectOwner)
			baseObjectData->propertyFlags=OBJECT_FLAG_CLIENT_OWNER;
		else
			baseObjectData->propertyFlags=OBJECT_FLAG_CLIENT_NON_OWNER;
	}
}
*/

void DataReplicator::OnCreationCommand(ReplicatedObject* newObject, PlayerID senderPlayerId, ObjectID objectId, PlayerID objectOwner, BaseObjectData *baseObjectData)
{
	if (newObject)
	{
		// If the user didn't set the network ID, set it now.
		if (newObject->GetNetworkID()==UNASSIGNED_OBJECT_ID)
			newObject->SetNetworkID(objectId);

		newObject->SetOwner(objectOwner);

		if (baseObjectData->calledOnReplication==false)
		{
			baseObjectData->calledOnReplication=true;
			newObject->OnReplication(this);
		}
	}
	else
	{
		SendObjectStopRequest(objectId,senderPlayerId);
		return;
	}
}
// Returns true to queue the packet, false otherwise
bool DataReplicator::ParsePacket(RakPeerInterface *peer, Packet *packet)
{

	unsigned char packetIdentifier;
	if ( ( unsigned char ) packet->data[ 0 ] == ID_TIMESTAMP )
	{
		if ( packet->length > sizeof( unsigned char ) + sizeof( unsigned int ) )
			packetIdentifier = ( unsigned char ) packet->data[ sizeof( unsigned char ) + sizeof( unsigned int ) ];
		else
			return 0;
	}
	else
		packetIdentifier = ( unsigned char ) packet->data[ 0 ];

	switch (packetIdentifier)
	{
	case ID_NEW_INCOMING_CONNECTION:
	case ID_CONNECTION_REQUEST_ACCEPTED:
		OnNewPlayer(peer, packet);
		return false;
	case ID_DISCONNECTION_NOTIFICATION:
	case ID_CONNECTION_LOST:
		OnLostPlayer(peer, packet);
		return false;
	case ID_REPLICATOR_STR_MAP_INDEX:
	//	printf("ID_REPLICATOR_STR_MAP_INDEX Used\n");
		OnStringMapIndex(peer, packet);
		return false;
	case ID_REPLICATOR_DATA_SEND_MEMORY:
	//	printf("ID_REPLICATOR_DATA_SEND_MEMORY");
		if (enableReplication==false)
		{
	//		printf(" Ignored\n");
			return true;
		}
		else
		{
	//		printf(" Used\n");
			OnDataReplicateSendMemory(peer, packet);
		}
		return false;
	case ID_REPLICATOR_DATA_PUSH_OBJECT:
//		printf("ID_REPLICATOR_DATA_PUSH_OBJECT");
		if (enableReplication==false)
		{

	//		printf(" Ignored\n");
			return true;
		}
		else
		{
		//	printf(" Used\n");
			objectCreatedFromNetwork=true;
			OnSendReplicationPushPacket(peer, packet);
			objectCreatedFromNetwork=false;
		}
		return false;
	case ID_REPLICATOR_DATA_SEND_OBJECT_SCOPE:
//		printf("ID_REPLICATOR_DATA_SEND_OBJECT_SCOPE");
		if (enableReplication==false)
		{
		//	printf(" Ignored\n");
			return true;
		}
		else
		{
		//	printf(" Used\n");
			OnDataReplicateSendObjectScope(peer, packet);
		}
		return false;
		/*
	case ID_REPLICATOR_MEMORY_START_REQUEST:
		printf("ID_REPLICATOR_MEMORY_START_REQUEST");
		if (enableReplication==false)
		{
			printf(" Ignored\n");
			return true;
		}
		else
		{
			printf(" Used\n");
			OnMemoryReplicateStart(peer, packet);
		}
		return false;
		*/
	case ID_REPLICATOR_DATA_STOP:
	//	printf("ID_REPLICATOR_DATA_STOP");
		if (enableReplication==false)
		{
	//		printf(" Ignored\n");
			return true;
		}
		else
		{
	//		printf(" Used\n");
			OnDataReplicateStop(peer, packet);
		}
		return false;
	case ID_REPLICATOR_OBJECT_CREATION_REQUEST:
//		printf("ID_REPLICATOR_OBJECT_CREATION_REQUEST");
		if (enableReplication==false)
		{
		//	printf(" Ignored\n");
			return true;
		}
		else
		{
		//	printf(" Used\n");
			objectCreatedFromNetwork=true;
			OnObjectCreationRequest(peer, packet);
			objectCreatedFromNetwork=false;
		}
		return false;
	case ID_REPLICATOR_PUSH_COMPLETE:
		//		printf("ID_REPLICATOR_PUSH_COMPLETE");
		if (enableReplication==false)
		{
			//	printf(" Ignored\n");
			return true;
		}
		else
		{
			//	printf(" Used\n");
			OnPushComplete(peer, packet);
		}
		return false;
	case ID_REPLICATOR_OBJECT_CREATION_REQUEST_RESPONSE:
	//	printf("ID_REPLICATOR_OBJECT_CREATION_REQUEST_RESPONSE");
		if (enableReplication==false)
		{
		//	printf(" Ignored\n");
			return true;
		}
		else
		{
	//		printf(" Used\n");
			objectCreatedFromNetwork=true;
			OnObjectCreationRequestResponse(peer, packet);
			objectCreatedFromNetwork=false;
		}
		return false;
	}

	return false;
}
int DataReplicator::OnReceive(RakPeerInterface *peer, Packet *packet)
{
	assert(packet);
	assert(peer);

	Packet *queuedPacket;
	// Process and deallocate queued packets
	while (packetQueue.size() && enableReplication)
	{
		queuedPacket=packetQueue.pop();
		ParsePacket(peer, queuedPacket);
		delete queuedPacket->data;
		delete queuedPacket;
	}

	if (ParsePacket(peer, packet))
	{
		// Queue a copy of the packet
		Packet *copy;
		copy = new Packet;
		copy->bitSize=packet->bitSize;
		copy->data=new unsigned char [packet->length];
		memcpy(copy->data, packet->data, packet->length);
		copy->deletePacket=packet->deletePacket;
		copy->length=packet->length;
		copy->playerId=packet->playerId;
		copy->playerIndex=packet->playerIndex;
		packetQueue.push(copy);
		return 1; // We absorbed the packet, the caller will deallocate it
	}

	// If we do not propagate this kind of packet to the game, return true to signal that this class absorbed the packet
	if (PropagateToGame(packet)==false)
		return 1; // We absorbed the packet, the caller will deallocate it
	else
		return 0; // Used the packet, but didn't absorb it
}
void DataReplicator::OnDisconnect(RakPeerInterface *peer)
{
}
bool DataReplicator::PropagateToGame(Packet *packet) const
{
	unsigned char packetIdentifier;
	if ( ( unsigned char ) packet->data[ 0 ] == ID_TIMESTAMP )
	{
		if ( packet->length > sizeof( unsigned char ) + sizeof( unsigned int ) )
			packetIdentifier = ( unsigned char ) packet->data[ sizeof( unsigned char ) + sizeof( unsigned int ) ];
		else
			return true;
	}
	else
		packetIdentifier = ( unsigned char ) packet->data[ 0 ];

	return packetIdentifier!=ID_REPLICATOR_DATA_PUSH_OBJECT &&
		packetIdentifier!=ID_REPLICATOR_DATA_SEND_MEMORY &&
		packetIdentifier!=ID_REPLICATOR_DATA_SEND_OBJECT_SCOPE &&
		//packetIdentifier!=ID_REPLICATOR_MEMORY_START_REQUEST &&
		packetIdentifier!=ID_REPLICATOR_DATA_STOP &&
		packetIdentifier!=ID_REPLICATOR_OBJECT_CREATION_REQUEST &&
		packetIdentifier!=ID_REPLICATOR_PUSH_COMPLETE &&
		packetIdentifier!=ID_REPLICATOR_OBJECT_CREATION_REQUEST_RESPONSE &&
		packetIdentifier!=ID_REPLICATOR_STR_MAP_INDEX;
}
unsigned DataReplicator::ParticipantStruct::GetExtendedMemoryIndexByKey(strToIndexMapType key)
{
	unsigned index;
	for (index=0; index < extendedMemoryList.Size(); index++)
	{
		if (extendedMemoryList[index]->remoteKey==key)
			return index;
	}
	return (unsigned)-1;
}
DataReplicator::BaseMemoryData::BaseMemoryData()
{
	localIdentifier=0;
}
DataReplicator::BaseMemoryData::~BaseMemoryData()
{
	if (localIdentifier)
		delete [] localIdentifier;
}

DataReplicator::ParticipantStruct::ParticipantStruct()
{
	objectRequestCount=0;
}
DataReplicator::ParticipantStruct::~ParticipantStruct()
{
	unsigned i;
	for (i=0; i < extendedMemoryList.Size(); i++)
		delete extendedMemoryList[i];
	for (i=0; i < extendedObjectList.Size(); i++)
		delete extendedObjectList[i];
	for (i=0; i < remoteObjectNameTable.size(); i++)
		delete remoteObjectNameTable[i];
	for (i=0; i < remoteMemoryNameTable.size(); i++)
		delete remoteMemoryNameTable[i];
	extendedMemoryList.Clear();
	extendedObjectList.Clear();
	remoteObjectNameTable.clear();
	remoteMemoryNameTable.clear();
}

ReplicatedObject::ReplicatedObject()
{
	owner=UNASSIGNED_PLAYER_ID;
}
ReplicatedObject::~ReplicatedObject()
{
}
bool ReplicatedObject::InScope(PlayerID playerId, DataReplicator *dataReplicator)
{
	return true;
}
void ReplicatedObject::OnScopeChange(PlayerID sender, bool isNowInScope, DataReplicator *dataReplicator)
{
}
bool ReplicatedObject::AcceptDereplicationRequest(PlayerID sender, DataReplicator *dataReplicator)
{
	return true;
}
void ReplicatedObject::OnRemoteDereplication(DataReplicator *dataReplicator)
{
	delete this;
}
void ReplicatedObject::OnReplication(DataReplicator *dataReplicator)
{
	// Synchronize member variables here
}
bool ReplicatedObject::OnReceiveCreationRequest(PlayerID sender, RakNet::BitStream *inContext, RakNet::BitStream *outContext, PlayerID &objectOwner, DataReplicator *dataReplicator)
{
	// Return true to allow the object to be created.
	return true;
}
void ReplicatedObject::SetOwner(PlayerID _owner)
{
	owner = _owner;
}
PlayerID ReplicatedObject::GetOwner(void) const
{
	return owner;
}
bool ReplicatedObject::GetWaitingForCreationResponse(void)
{
	return GetNetworkID()==UNASSIGNED_OBJECT_ID;
}
bool ReplicatedObject::OnCreationRequestAcceptance(PlayerID sender, ObjectID newObjectId, RakNet::BitStream *inContext, PlayerID &objectOwner, DataReplicator *dataReplicator)
{
	return true;
}
void ReplicatedObject::OnCreationRequestRejection(PlayerID sender, RakNet::BitStream *inContext, DataReplicator *dataReplicator)
{
}
ObjectReplicationPushResult ReplicatedObject::OnSendCreationRequest(PlayerID destination, RakNet::BitStream *outContext, DataReplicator *dataReplicator)
{
	return OBJECT_REPLICATION_OK;
}
ObjectReplicationPushResult ReplicatedObject::OnSendReplicationPush(PlayerID destination, RakNet::BitStream *outContext, DataReplicator *dataReplicator)
{
	return OBJECT_REPLICATION_OK;
}
bool ReplicatedObject::OnReceiveReplicationPush(PlayerID sender, RakNet::BitStream *inContext, ObjectID newObjectId, PlayerID &objectOwner, DataReplicator *dataReplicator)
{
	return true;
}
float DataReplicator::GetCumulativeInterpolationPercentile(unsigned int currentTimeMS, unsigned int timePacketSentMS, unsigned int lastCallTimeMS, unsigned int interpolationTimeMS)
{
    unsigned int finalTime = timePacketSentMS + interpolationTimeMS;
	if (currentTimeMS < timePacketSentMS)
		return 0.0f;
	if (lastCallTimeMS >= finalTime)
		return 1.0f;
	if (currentTimeMS < finalTime)
		return (float)(currentTimeMS - lastCallTimeMS) / (float)(finalTime - lastCallTimeMS);
	else
		return 1.0f;
}

float DataReplicator::GetLinearInterpolationPercentile(unsigned int currentTimeMS, unsigned int timePacketSentMS, unsigned int lastCallTimeMS, unsigned int interpolationTimeMS)
{
	unsigned int finalTime = timePacketSentMS + interpolationTimeMS;
	if (currentTimeMS <= timePacketSentMS)
		return 0.0000001f; // Should be 0, but 0 is used to signal that interpolation is done
	if (lastCallTimeMS >= finalTime)
		return 0.0f;
    if (currentTimeMS <= finalTime)
		return (float)(currentTimeMS - lastCallTimeMS) / (float)interpolationTimeMS;
	else
		return (float)(finalTime - lastCallTimeMS) / (float)interpolationTimeMS;
}

// Returns true if we finished updating position, false if we did not
bool DataReplicator::InterpolateFloat(float *position, float targetOrigin, float targetVelocity, unsigned int currentTimeMS, unsigned int timePacketSentMS, unsigned int lastCallTimeMS, unsigned int interpolationTimeMS, bool updatePastInterpolationTime)
{
	if (currentTimeMS - timePacketSentMS < interpolationTimeMS)
	{
		float targetFinalPosition;
		targetFinalPosition=targetOrigin + targetVelocity * interpolationTimeMS;
		// Get original origin from last position and time
		// pos = origin + (final_pos - origin) * elapsed_percentile
		// pos = origin + final_pos * elapsed_percentile - origin * elapsed_percentile
		// pos - (final_pos * elapsed_percentile) = origin - origin * elapsed_percentile
		// pos - (final_pos * elapsed_percentile) = origin * ( 1 - elapsed_percentile)
		// (pos - final_pos * elapsed_percentile) / (1 - elapsed_percentile) = origin

		float lastElaspedPercentile = (float)(lastCallTimeMS - timePacketSentMS) / interpolationTimeMS;
		float currentElaspedPercentile = (float)(currentTimeMS - timePacketSentMS) / interpolationTimeMS;
		float origin;

		// Invert the last calculation to get the origin
		if (lastElaspedPercentile > 0.0f)
			origin = (*position - targetFinalPosition * lastElaspedPercentile) / (1.0f - lastElaspedPercentile);
		else
			origin = *position;

		// Now given the origin, find out where we should be at the current time
		*position = origin + (targetFinalPosition - origin ) * currentElaspedPercentile;
	}
	else
	{
		if (targetVelocity==0.0f)
		{
			*position=targetOrigin;
			return true;
		}

		*position=targetOrigin + targetVelocity * (float)(lastCallTimeMS - timePacketSentMS);

		if (updatePastInterpolationTime==false)
			return true;
	}

	return false;
}
